% clear all;
% close all;
% addpath(genpath('/home/armand/MATLAB/CODE_PAPER/PCA_CODE_SOURCE'));
%
% Copyright (c) 2011, Ludwig Institute for Cancer Research.
% All rights reserved.
% Author: Armand Valsesia <armand.valsesia@licr.org>
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%     * Redistributions of source code must retain the above copyright
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above
%       copyright notice, this list of conditions and the following
%       disclaimer in the documentation and/or other materials
%       provided with the distribution.
%     * Neither the name of the <organization> nor the
%       names of its contributors may be used to endorse or promote
%       products derived from this software without specific prior
%       written permission.
%
% THIS SOFTWARE IS PROVIDED BY GRL ``AS IS'' AND ANY EXPRESS OR
% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED.  IN NO EVENT SHALL GRL BE LIABLE FOR ANY DIRECT,
% INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
% HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
% STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
% OF THE POSSIBILITY OF SUCH DAMAGE.
%
%
% ALGORITHM OVERVIEW :
%   - convert input file to .mat
%   - perform simpleMerge
%   - segment genome into regions separated by long monomorphic regions
%   - foreach regions :
%       - do a PCA on SNP*sample matrix
%       - use main PCs to train a Self-Organizing Map (SOM)
%       - cluster SNPs using the trained SOM
%       - merge into regions, adjacent SNPs from a same cluster
%%%%%%%%%%%%%%%%
function PCAmerge(varargin)

% Defining property parameters and default values.

par.input=NaN; %'/home/armand/MATLAB/CODE_PAPER/GMM_CNV_outputs/test.dat.cnv.dat';

par.delim='\t'; % column delimiter in input

par.maxDist=500000; % max distance allowed when merging adjacent SNPs (in Simple Merge)

par.W = 1; % partition genome by finding long strech of monomorphic regions, having a length (in Mb) >= W

par.varExp=0.90;% of variance explained by main PCs

par.clean=0; % remove everything except the final .txt file

par.outputDir=NaN; %'/tmp/PCAtest/'; % root output directory

par.plot=1;


%Parsing parameters
par=parsepvpairsparstruct(par,varargin{:});

% ensure args are double not string
par.maxDist=str2double(par.maxDist);
par.W =str2double(par.W);
par.varExp=str2double(par.varExp);
par.clean=str2double(par.clean);
par.plot=str2double(par.plot);

if exist(par.input,'file')==0
    error('No or invalid input file specified')
end

if par.plot==1 && par.clean==1
    error('Warning option clean and plot are both turned on.\nNot sure about your intentions...\nAborting')
end

mymkdir(par.outputDir);



%% convert file to .mat
% return CHR : vector of chr nb (with Mt = 0, X = 23 and Y = 24)
% POS : vector of SNP pos, SAMPLES : vector of samples IDs
% DATA : matrix SNP * SAMPLES, where element is copy number value
[CHR POS DATA SAMPLES]=convert_tab2mat(par.input, par.delim, [par.outputDir 'master_input.mat']);

%% simple merge
par.OUTPUT_SIMPLE_MERGE=[par.outputDir 'simpleMerge/'];
mymkdir(par.OUTPUT_SIMPLE_MERGE);
chr_list=simple_merge([par.outputDir 'master_input.mat'],par.OUTPUT_SIMPLE_MERGE,par.maxDist);

%% segment genome into smaller piece to investigate
INITIAL_REGIONS =segment_genome([par.outputDir 'master_input.mat'],[par.outputDir 'regions4pca.mat'],par.W,par.OUTPUT_SIMPLE_MERGE,chr_list);


%% iterate genomic regions in INITIAL_REGIONS
%% and perform PCA merge
par.OUTPUT_PCA_MERGE=[par.outputDir 'pcaMerge/'];
mymkdir(par.OUTPUT_PCA_MERGE)
MERGED_REGIONS=[];
nInds=length(SAMPLES);

last_id=0;
for i=1:length(chr_list)% iterate chr
    
    % get data for this chr
    chr=chr_list(i);
    chr_ix=find(INITIAL_REGIONS(:,1) == chr);
    
    % iterate regions of interest for this chr
    for j=1:length(chr_ix)
        
        disp(['doing ', sprintf(' - region chr%d:%d-%d',chr,INITIAL_REGIONS(chr_ix(j),2),INITIAL_REGIONS(chr_ix(j),3))])
        
        % set up output dir
        output_dir = [par.OUTPUT_PCA_MERGE ...
            '/chr' num2str(chr) '_' num2str(INITIAL_REGIONS(chr_ix(j),2)) '_' num2str(INITIAL_REGIONS(chr_ix(j),3)) '/' ];
        mymkdir(output_dir);
        
        try
            load( [output_dir 'merged_regions.mat'])
            disp('reloading data');
            
        catch
            
            %% subset data
            
            % get SNPs in this region of interest
            sel = find( CHR == chr & POS >= INITIAL_REGIONS(chr_ix(j),2) & POS <= INITIAL_REGIONS(chr_ix(j),3));
            pos=POS(sel);
            D=DATA(sel,:);
            
            % calc CNV frequency in this region
            freq=100* sum(round(D')~=2) /nInds;
            
            
            
            %%
            %% PCA on this genomic region
            
            % do PCA straight on localSNP.data, note it will center DATA by
            % substract off columns mean but does not rescale the column
            % of DATA
            fprintf('Doing PCA\n')
            [COEFF,SCORE, latent ]=princomp(D','econ');
            useful_pc= find(latent > 0);
            if isempty(useful_pc)
                continue
            end
            V=COEFF(:,useful_pc);
            d=latent(useful_pc);
            
            
            
            
            %% select PCs explaining most of the variance
            variance_explained=cumsum(d)/sum(d);
            higher_cluster_indexes=find( variance_explained <= par.varExp );
            if isempty(higher_cluster_indexes)
                higher_cluster_indexes=1; % special case where 1st PC explains 100% of the variance
            end
            
            % set the number of component to use for clustering the data
            % into regions
            nPCs2use=length(higher_cluster_indexes);
            
            if nPCs2use==1
                nPCs2use=nPCs2use+1;
            end
            Vrescaled= V.*repmat(d', size(V,1), 1);
            X=Vrescaled(:,1:nPCs2use)';
            
            
            %             % plot PCs
            %             subplot(2,1,1)
            %             imagesc(V')
            %             ylabel('PCs');
            %             xlabel('Markers');
            %             colorbar;
            %             title(sprintf('CNV region : chr%d:%d-%d',chr,INITIAL_REGIONS(chr_ix(j),2),INITIAL_REGIONS(chr_ix(j),3)))
            %             % plot lambdas spectrum
            %             subplot(2,1,2)
            %             hold on
            %              plot(d,'b.','MarkerSize',12)
            %             plot(d(higher_cluster_indexes),'r.','MarkerSize',12)
            %               hold off
            %             xlabel('PCs')
            %             ylabel('\lambda')
            %             title('Eigen values');
            %             legend('cluster1 : big \lambda','cluster2 small \lambda','Location','Best');
            %
            % imagesc(X)
            
            %% Self-Organizing Map
            fprintf('Training SOM on data\n')
            net = newsom(minmax(X),[5 3]);
            % train neurones on main PCs
            net = train(net, X);
            % cluster data
            fprintf('SOM clustering\n')
            
            distances = dist(X',net.IW{1}');
            [dummy,idx] = min(distances,[],2);
            
             %% ensure cluster idx are unique genome wide
            idx=idx+last_id;
            last_id=last_id + max(idx);
          
            %% REMERGE REGIONS from the latest clustering
            
            fprintf('Merging into regions\n')
            % create regions (with start,end = CNV index) made of CNVs belonging to a
            % same cluster
            regions=find_adjacent_regions(idx);
            
            
            merged_genomic_regions=create_region_from_snps(regions,pos,chr, freq);
            %[ merged_genomic_regions(:,1:2) merged_genomic_regions(:,3:4)/1e6 merged_genomic_regions(:,5:end)]
            
            
            if par.plot ==1
                % plot main PCs
                subplot(3,1,1)
                imagesc(D')
                colorbar;
                ylabel('individuals')
                title('CN values')
                
                subplot(3,1,2)
                imagesc(X)
                ylabel([num2str(nPCs2use) 'st PCs']);
                colorbar;
                set(gca,'YTick',1:nPCs2use);
                set(gca,'YTickLabel',1:nPCs2use);
                %             title(sprintf('CNV region : chr%d:%d-%d',chr,INITIAL_REGIONS(chr_ix(j),2),INITIAL_REGIONS(chr_ix(j),3)))
                title('main PCs')
                subplot(3,1,3)
                % plot regions
                for ii=1:size(merged_genomic_regions,1) % plot new regions
                    
                    %                  hLine2 =  line([merged_genomic_regions(ii,3)/1e6 merged_genomic_regions(ii,3)/1e6], [merged_genomic_regions(ii,7) merged_genomic_regions(ii,7)],...
                    %                     'Color', 'blue',...
                    %                     'LineWidth', 1.5);
                    hLine2 =  line([regions(ii,1) regions(ii,2)], [merged_genomic_regions(ii,7) merged_genomic_regions(ii,7)],...
                        'Color', 'blue',...
                        'LineWidth', 3);
                    hold on
                    if ii > 1
                        set(get(get(hLine2,'Annotation'),'LegendInformation'),...
                            'IconDisplayStyle','off'); % Exclude line from legend
                    end
                end
                xlabel('SNP index');
                ylabel('CNV frequency (%)');
                title('PCA merged regions')
                % save fig
                %             hgsave([myoutput 'merged_regions.fig'])
                saveas(gcf,[output_dir 'merged_regions'], 'pdf')
                close
            end
            
            % save data to .mat
            % V = PCA matrix SNPs * all PCs
            % d = eigen vector
            % nPC2use : nb of main PCs, can be used to retrieve SNPs *
            % mainPC matrix from V and d, which was used for the SOM
            % net = SOM object, after training,
            % merged_genomic_regions = table with merged regions, contains :
            %   cluster_id, chr, start, end, #SNPs, minFreq, medianFreq (fix), maxFreq
            save( [output_dir 'merged_regions.mat'], 'merged_genomic_regions', 'V','d','net','nPCs2use' )
            
        end
        
        MERGED_REGIONS = [ MERGED_REGIONS ; merged_genomic_regions];
        clear merged_genomic_regions
    end % end of iteration for genomic window (on a given chr)
    
end % end of iteration for chr

% save merged_genomic_regions to txt file
fprintf('Saving data to %s\n',[par.outputDir 'merged_regions.txt']);
print_report( [par.outputDir 'merged_regions.txt'] ,...
    {'clusterId','chr','start','end','length','minFreq','medianFreq','maxFreq'}, ...
    {  MERGED_REGIONS(:,1), MERGED_REGIONS(:,2), MERGED_REGIONS(:,3), MERGED_REGIONS(:,4), MERGED_REGIONS(:,5), MERGED_REGIONS(:,6), MERGED_REGIONS(:,7), MERGED_REGIONS(:,8) }, ...
    [repmat(2, 1, 5) repmat(1,1,3) ] ,1);
%%
if par.clean ==1
    % remove everything except final [outputDir 'merged_regions.txt']
    delete([par.outputDir 'master_input.mat'])
    delete([par.outputDir 'regions4pca.mat'])
    rmdir(par.OUTPUT_PCA_MERGE,'s')
    rmdir(par.OUTPUT_SIMPLE_MERGE,'s')
end
%%
end
