% Copyright (c) 2011, Ludwig Institute for Cancer Research.
% All rights reserved.
% Author: Armand Valsesia <armand.valsesia@licr.org>
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%     * Redistributions of source code must retain the above copyright
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above
%       copyright notice, this list of conditions and the following
%       disclaimer in the documentation and/or other materials
%       provided with the distribution.
%     * Neither the name of the <organization> nor the
%       names of its contributors may be used to endorse or promote
%       products derived from this software without specific prior
%       written permission.
%
% THIS SOFTWARE IS PROVIDED BY GRL ``AS IS'' AND ANY EXPRESS OR
% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED.  IN NO EVENT SHALL GRL BE LIABLE FOR ANY DIRECT,
% INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
% HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
% STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
% OF THE POSSIBILITY OF SUCH DAMAGE.
%
%
% function simple_merge
%   merge adjacent SNPs having strictly the same CNV profile
%
%   argins :
%       - input file as generated by convert_tab2mat
%       - ouput directory
%       - maxDistSNP : max distance allowed between 2 adjacent SNPs
%
%   results will be saved to :
%       - [file .mat], a Matlab binary file, containing
%               'regions' : a table listing all merged regions, columns are
%               chr start end #SNPs #individuals having a CNV
%               'regions_data' : matrix of samples * regions, where element
%               is the CN value
%               'snp_cnvFreq' : for each SNP, # of individuals where its
%               SNP was found variant (deleted or duplicated)
%       - [file .txt] a flat file containing the 'regions' table (cf above)
%
function chr_list=simple_merge(file,outputDIR,  maxDistSNP )

if nargin < 1
    error('Simple_merge : No input file specified')
end
if nargin < 2
    error('Simple_merge : No output directory specified')
end
mymkdir(outputDIR);

% check/set max distance allowed between 2 adjacent SNPs
% in a same region
if nargin < 3
    maxDistSNP=500000;
    disp(['  setting maxDistSNP to ',num2str(maxDistSNP)]);
end

mymkdir(outputDIR);

% margin = number of differences allowed between SNPs in order to merge
% them in a single region
margin=0;

% load chr : vector of chr nb (with Mt = 0, X = 23 and Y = 24)
% pos : vector of SNP pos, samples : vector of samples IDs
% data : matrix SNP * SAMPLES, where element is copy number value
load(file ,'chr','pos','data','samples')

data=round(data); % ensure we work with discrete CN values
% fprintf('%s %d %d\n','dataO',size(data) )
% work by chr
chr_list=unique(chr);
for i=1:length(chr_list)
    fprintf('Simple Merge chr%d\n', chr_list(i))

    sel=chr==chr_list(i);
    output=[ outputDIR 'chr' num2str(chr_list(i)) ];
    try
        load(strcat(output, '.mat'))
        fprintf('Reloading data from %s\n',output)
    catch
        
        do_simple_merge( output ,  chr_list(i), pos(sel), data(sel,:)', margin, maxDistSNP);
    end
end

end

function do_simple_merge( output , chr, pos, data, margin, maxDistSNP)

%% calc diff between columns
fprintf('\t%s\n','Merging into regions');
% fprintf('%s %d %d\n','data',size(data) )
diff = reshape( sum(abs(data(:,1:end-1) - data(:,2:end) ) ) ,[],1);
% fprintf('%s %d %d\n','diff',size(diff) )
% calc distance between adjacent SNPs
dist = reshape( abs(pos(1:(end-1)) - pos(2:end) ) ,[],1);
% fprintf('%s %d %d\n','dist',size(dist) )
%% define regions

% get indexes where col i and col i+1 are different enough
% or where dist between SNPs is > maxallowed
%disp('Getting the big-enough differences');

indexes = find(diff > margin | dist > maxDistSNP);

% get real indexes for these SNPs ( get intervals)
real_SNPindexes = [1; indexes + 1];

% get matrix of region's SNPs
regions_data = data( :, real_SNPindexes );

regionId=1;
regions =[]; %zeros( length(real_SNPindexes),4);
regionIds=zeros(length(pos),1);
%% define regions
if ~isempty(indexes)==1
    % iterate indexes (mapping to columns that differ from their next
    % neighbour)
    % IMPORTANT :
    % differences are counted between SNPi and SNPi+1
    % so the index in diff - correspond to diff(SNPi, SNPi+1)
    % when an element e(j) at index j in vector diff is > threshold
    % it means that a region is ending at SNP(e(j)) and a new one starts at
    % SNP(e(j)+1)
    %    disp('differences found');

    % save first region
    if is_monomorphic(data(:,1:indexes(1)))==1
        % store this monomorphic region for later
        % the reason is the distance condition will break most of
        % monomorphic regions into small segments. but those need to be fusion
        % together
        monomorphicR_to_save=[1,indexes(1)];

    else
        % this is a polymorphic region, we can save it now
        % without worrying about this fusion of monoM regions
        regions = [regions ; ...
            [regionId, chr , min( pos( 1:indexes(1) ) ),max( pos( 1:indexes(1) ) ),         length(pos( 1:indexes(1) )) ,   sum(data(:,1)~=2) ] ];
        %             create_region( data(:,1:indexes(1)), chr, regionId ,pos( 1:indexes(1) ) ) ];
        regionIds(1:indexes(1)) =regionId;
        regionId=regionId+1;
        monomorphicR_to_save=NaN;

    end
    nIter=length(indexes);

    fprintf('\tfound %d regions\t\n',nIter+1);
    % iterate next regions
    for i=1:length(indexes)-1
        % iterate vector of "significant" differences

        % notation is a bit confusing here , so as a remainder :
        % i is the index in vector indexes
        % element indexes(i) is the difference SNPk, SNPk+1, so indexes(i)
        % = index in data of SNPk
        if(~mod(i-1,100))
            fprintf('.'); % print advancement every 100 regions
        end

        index = indexes(i);

        if is_monomorphic(data(:,(index+1):indexes(i+1)))==1
            % this region is monomorphic, check if it is adjacent to any
            % previous monoM region
            if ~isnan(sum(monomorphicR_to_save))==1
                % fusion with previous monoM region
                monomorphicR_to_save(2)=indexes(i+1);
            else
                % no previous monoM region
                monomorphicR_to_save=[index+1,indexes(i+1)];
            end
        else
            if ~isnan(sum(monomorphicR_to_save))==1
                % save previous monomorphic region
                regions = [regions ; ...
                    [regionId, chr , min( pos( monomorphicR_to_save(1):monomorphicR_to_save(2) )  ),max( pos( monomorphicR_to_save(1):monomorphicR_to_save(2) )  ),   ...
                    length(pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ) ,   sum(data(:,monomorphicR_to_save(1))~=2) ] ];

                %                     create_region( data(:,monomorphicR_to_save(1):monomorphicR_to_save(2)), chr, ...
                %                     regionId ,pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ) ];
                regionIds(monomorphicR_to_save(1):monomorphicR_to_save(2)) =regionId;
                regionId=regionId+1;
                monomorphicR_to_save=NaN;
            end


            % save this region
            regions = [regions ; ...
                [regionId, chr , min( pos( (index+1):indexes(i+1) ) ),max( pos( (index+1):indexes(i+1) ) ),         length(pos( (index+1):indexes(i+1) )) ,   sum(data(:,index+1)~=2) ]];
            %                 create_region( data(:,(index+1):indexes(i+1)), chr, regionId ,pos( (index+1):indexes(i+1) ) ) ];
            regionIds((index+1):indexes(i+1)) =regionId;
            regionId=regionId+1;
            monomorphicR_to_save=NaN;
        end %end is_monomorphic(data(:,(index+1):indexes(i+1)))
    end % end for i=1:length(indexes)-1

    % need to save last region
    if(indexes(end) ~= size(data, 2) )
        % no differences found foreach SNPs between between SNP1 and SNPn ( with n = indexes(1) )
        % we need to create the appropriate regions

        lastI=size(data, 2);
        if is_monomorphic(data(:,indexes(end)+1:lastI))==1
            % this region is monomorphic, check if it is adjacent to any
            % previous monoM region
            if ~isnan(sum(monomorphicR_to_save))
                % fusion with previous monoM region
                monomorphicR_to_save(2)=lastI;
            else
                % no previous monoM region
                monomorphicR_to_save=[indexes(end)+1,lastI];
            end

            % save it as no further SNPs
            regions = [regions ; ...
                [regionId, chr , min( pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ),max( pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ),    ...
                length(pos( monomorphicR_to_save(1):monomorphicR_to_save(2) )) ,   sum(data(:,monomorphicR_to_save(1))~=2) ] ];
            %                 create_region( data(:,monomorphicR_to_save(1):monomorphicR_to_save(2)), chr, ...
            %                 regionId ,pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ) ];

            regionIds(monomorphicR_to_save(1):monomorphicR_to_save(2)) =regionId;
            regionId=regionId+1;

            monomorphicR_to_save=NaN;

        else
            if  ~isnan(sum(monomorphicR_to_save))==1
                % save previous monomorphic region
                regions = [regions ; ...
                    [regionId, chr , min( pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ),max( pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ),  ...
                    length(pos( monomorphicR_to_save(1):monomorphicR_to_save(2) )) ,   sum(data(:,monomorphicR_to_save(1))~=2) ] ];
                %                     create_region( data(:,monomorphicR_to_save(1):monomorphicR_to_save(2)), chr, ...
                %                     regionId ,pos( monomorphicR_to_save(1):monomorphicR_to_save(2) ) ) ];

                regionIds(monomorphicR_to_save(1):monomorphicR_to_save(2)) =regionId;
                regionId=regionId+1;
                monomorphicR_to_save=NaN;
            end


            % save this region
            regions = [regions ; ...
                [regionId, chr , min( pos( indexes(end)+1:lastI ) ),max( pos( indexes(end)+1:lastI ) ),   ...
                length(pos( indexes(end)+1:lastI )) ,   sum(data(:,indexes(end))~=2) ] ];

            %                 create_region( data(:,indexes(end)+1:lastI), chr, regionId ,pos( indexes(end)+1:lastI ) ) ];

            regionIds((indexes(end)+1):lastI) =regionId;
            regionId=regionId+1;
            monomorphicR_to_save=NaN;
        end %end is_monomorphic(data(:,(index+1):indexes(i+1)))


    end %end  if(indexes(end) ~= size(data, 2) )

else
    % all columns are strictly identicals
    % this should be a pretty rare event - but being cautious is better
    % than rather not
    fprintf('\t%s\n','all columns from data are identicals !!');
    %      if(VERBOSE>1)
    %         disp(strcat('saving last from SNP ', num2str(1), ' to ',num2str(size(data, 2) ) ));
    %      end
    regions = [regions ; [regionId, chr , min( pos ),max( pos ),         length(pos) ,   sum(data(:,1)~=2) ] ];
end %end if(indexes)


%% calculating SNP MAF

fprintf('\t%s\n','calculating CNV frequency');
snp_cnvFreq=sum(data~=2);


%%
fprintf('\t%s\n','Writing outputs ');

labels={ 'chr',     'start',   'end',     'length',  'occurences'};
num=[2 2 2 2 1 ];
V={ regions(:,1);regions(:,2);regions(:,3);regions(:,4);regions(:,5)};
print_report([output '.txt'],labels,V,num);

% save session as mat file
save(strcat(output, '.mat'), 'regions' ,'regions_data','snp_cnvFreq' ,'-v7.3' );


end

